bl_info = {
    "name": "Real time paths",
    "author": "Hamdi",
    "description": "Automatically updates the motion path after a keyframe is moved",
    "blender": (4, 1, 0),
    "version": (1, 0, 0),
    "location": "Graph Editor ",
    "warning" : "this addon is for all devices and complex rigs",
    "category": "Animation",
}


import bpy


#### ------------------------------ UI ------------------------------ ####

class SNA_PT_MOTION_PATHS_882F1(bpy.types.Panel):
    bl_label = 'Motion Paths'
    bl_idname = 'SNA_PT_MOTION_PATHS_882F1'
    bl_space_type = 'GRAPH_EDITOR'
    bl_region_type = 'UI'
    bl_category = 'Real_Time_Paths'

    @classmethod
    def poll(cls, context):
        return True

    def draw(self, context):
        layout = self.layout

        wm = context.window_manager

        # Main Box that contains all
        main_box = layout.box()

        # Box for "Path Object" and "Path Bone"
        box_1 = main_box.box()
        row = box_1.row(align=False)
        row.operator('object.paths_calculate', text='Path Object', icon='OBJECT_DATA')
        row.operator('pose.paths_calculate', text='Path Bone', icon='BONE_DATA')

        # Box for "Delete Path", "Update Path", and "Delete All Paths"
        box_2 = main_box.box()
        row = box_2.row(align=False)
        row.operator('sna.delet_path_6e947', text='', icon='CANCEL')
        row.operator('object.paths_update_visible', text='Update Path', icon='FILE_REFRESH')
        row.operator('sna.delet_allpath_b6bbf', text='', icon='X')


        # Main Box that contains all elements
        main_box = layout.box()

        # Box for Update Mode selection
        box_update_mode = main_box.box()
        box_update_mode.label(text="Update Settings")
        box_update_mode.prop(wm.real_time_paths, "update_mode", text="Update Mode")
        
        # Additional option when 'TIMER' mode is selected
        if wm.real_time_paths.update_mode == 'TIMER':
            box_update_mode.prop(wm.real_time_paths, "auto_update_timer_interval", text="Timer Interval")

        # Box for activation/deactivation buttons
        box_activation = main_box.box()
        if not wm.real_time_paths.auto_update_active:
            box_activation.operator("amth.auto_update_activate", text="Activate Auto Update", icon='PLAY')
        else:
            box_activation.operator("amth.auto_update_deactivate", text="Deactivate Auto Update", icon='PAUSE')



#### ------------------------------ OPERATORS ------------------------------ ####

class SNA_OT_Delet_Path_6E947(bpy.types.Operator):
    bl_idname = "sna.delet_path_6e947"
    bl_label = "delet_path"
    bl_description = "delete all paths from selected bone or object"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        # Check if there is an active object and it is in pose mode
        if bpy.context.active_object and bpy.context.active_object.mode == 'POSE':
            # Clear paths in pose mode
            bpy.ops.pose.paths_clear(only_selected=True)
        # Check if there is an active object
        if bpy.context.active_object:
            # Clear paths in object mode   
            bpy.ops.object.paths_clear(only_selected=True)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


class SNA_OT_Delet_Allpath_B6Bbf(bpy.types.Operator):
    bl_idname = "sna.delet_allpath_b6bbf"
    bl_label = "Delet_All-path"
    bl_description = "delete all paths from the Scene  "
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        # Check if there is an active object and it is in pose mode
        if bpy.context.active_object and bpy.context.active_object.mode == 'POSE':
            # Clear paths in pose mode
            bpy.ops.pose.paths_clear(only_selected=False)
        # Check if there is an active object
        if bpy.context.active_object:
            # Clear paths in object mode
            bpy.ops.object.paths_clear(only_selected=False)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


class AMTH_OT_AutoUpdateMotionPaths(bpy.types.Operator):
    """Auto-update motion paths after keyframe or handle movement"""
    bl_idname = "amth.auto_update_motion_paths"
    bl_label = "Auto Update Motion Paths (Improved)"
    bl_options = {'REGISTER', 'UNDO'}

    _timer = None
    _last_active_keyframe_values = None  # Track the values and handles of the last active keyframe
    _needs_update = False  # Flag to indicate update needed

    @classmethod
    def poll(cls, context):
        return context.area.type == 'GRAPH_EDITOR'

    def invoke(self, context, event):
        wm = context.window_manager
        self._last_active_keyframe_values = self._get_active_keyframe_values(context)
        self._needs_update = False  # Reset update flag

        if wm.real_time_paths.update_mode == 'TIMER':
            self._timer = wm.event_timer_add(wm.real_time_paths.auto_update_timer_interval, window=context.window)
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context, event):
        wm = context.window_manager

        active_keyframe_values = self._get_active_keyframe_values(context)
        if wm.real_time_paths.update_mode == 'TIMER' and event.type == 'TIMER':
            if active_keyframe_values != self._last_active_keyframe_values:
                self._needs_update = True
        elif wm.real_time_paths.update_mode == 'EVENT':
            if active_keyframe_values != self._last_active_keyframe_values:
                self._needs_update = True

        if self._needs_update:
            try:
                bpy.ops.object.paths_update_visible()  # Update motion paths
            except Exception as e:
                print("Error updating motion paths:", e)
            self._last_active_keyframe_values = active_keyframe_values
            self._needs_update = False  # Reset flag after update

        if not wm.real_time_paths.auto_update_active:
            self.cancel(context)
            return {'CANCELLED'}

        return {'PASS_THROUGH'}

    def cancel(self, context):
        wm = context.window_manager
        if self._timer:
            wm.event_timer_remove(self._timer)
            self._timer = None

    def _get_active_keyframes(self, context):
        active_object = context.active_object
        active_action = active_object.animation_data.action if active_object and active_object.animation_data else None
        if not active_action:
            return []

        active_keyframes = []
        for fcurve in active_action.fcurves:
            for keyframe in fcurve.keyframe_points:
                if keyframe.select_control_point or keyframe.handle_left_type != 'AUTO' or keyframe.handle_right_type != 'AUTO':
                    active_keyframes.append(keyframe)
        return active_keyframes

    def _get_active_keyframe_values(self, context):
        active_keyframes = self._get_active_keyframes(context)
        if active_keyframes:
            return [(keyframe.co[0], keyframe.co[1], keyframe.handle_left[0], keyframe.handle_left[1], keyframe.handle_right[0], keyframe.handle_right[1]) for keyframe in active_keyframes]
        return None


class AMTH_OT_AutoUpdateActivate(bpy.types.Operator):
    bl_idname = "amth.auto_update_activate"
    bl_label = "Activate Auto Update"

    def execute(self, context):
        context.window_manager.real_time_paths.auto_update_active = True
        bpy.ops.amth.auto_update_motion_paths('INVOKE_DEFAULT')
        return {'FINISHED'}


class AMTH_OT_AutoUpdateDeactivate(bpy.types.Operator):
    bl_idname = "amth.auto_update_deactivate"
    bl_label = "Deactivate Auto Update"

    def execute(self, context):
        context.window_manager.real_time_paths.auto_update_active = False
        return {'FINISHED'}



#### ------------------------------ PROPERTIES ------------------------------ ####

class ANIM_PG_real_time_paths(bpy.types.PropertyGroup):
    # RunTime properties

    auto_update_active: bpy.props.BoolProperty(
        default=False,
    )
    auto_update_timer_interval: bpy.props.FloatProperty(
        name="Timer Interval",
        description="Interval in seconds for the auto-update timer",
        default=0.1,
        min=0.1,
        max=1.0,
    )
    update_mode: bpy.props.EnumProperty(
        name="Update Mode",
        description="Choose between timer-based and event-based update modes",
        items=[
            ('TIMER', "Real_Time", "Update motion paths in real time Suitable for medium or powerful devices"),
            ('EVENT', "After_Change", "Update motion paths after moving the keyframes Suitable for weak devices")
        ],
        default='EVENT',
    )


#### ------------------------------ REGISTRATION ------------------------------ ####

classes = [
    SNA_PT_MOTION_PATHS_882F1,
    SNA_OT_Delet_Path_6E947,
    SNA_OT_Delet_Allpath_B6Bbf,
    AMTH_OT_AutoUpdateMotionPaths,
    AMTH_OT_AutoUpdateActivate,
    AMTH_OT_AutoUpdateDeactivate,
    ANIM_PG_real_time_paths,
]

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy.types.WindowManager.real_time_paths = bpy.props.PointerProperty(type=ANIM_PG_real_time_paths, name="Real Time Paths")


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    del bpy.types.WindowManager.real_time_paths


if __name__ == "__main__":
    register()
